from time import time
from typing import Callable


def timeit(n: int, func: Callable, *args, **kwargs) -> float:
    times = list()
    for _ in range(n):
        start = time()
        func(*args, **kwargs)
        times.append(time() - start)
    return sum(times) / n



from unittest.mock import Mock

from stepizer.loader import Loader
from stepizer.proxy import DefaultProxy
from stepizer.step import Step
from tests.unit import UnitCase


class TestStepProperties(UnitCase):
    def setUp(self) -> None:
        self.callable = Mock(__name__='Mock')
        self.step = Step(self.callable)

    def test_callable(self) -> None:
        function = self.step.callable

        self.assertIs(function, self.callable)

    def test_name(self) -> None:
        name = self.step.name

        self.assertIsInstance(name, str)
        self.assertEqual(self.callable.__name__, name)

    def test_loader(self) -> None:
        loader = self.step.loader

        self.assertIsInstance(loader, Loader)

    def test_proxy_class(self) -> None:
        proxy_class = self.step.proxy_class

        self.assertIs(proxy_class, DefaultProxy)

    def test_map_args(self) -> None:
        map_args = self.step.map_args

        self.assertIsInstance(map_args, tuple)

    def test_map_kwargs(self) -> None:
        map_kwargs = self.step.map_kwargs

        self.assertIsInstance(map_kwargs, dict)

    def test_next_step(self) -> None:
        next_step = self.step.next_step

        self.assertIsNone(next_step)



from typing import Any, Iterable

from stepizer.step import Step
from tests.benchmark import BenchmarkCase
from tests.benchmark.utils import timeit


def function(prefix: str, number: int = 5) -> Iterable[str]:
    for i in range(number):
        x = sum([a ** i for a in range(1000)])
        yield f"{prefix}.{x}-{i}"


def baseline_pipeline(*args, **kwargs) -> Any:
    out = []
    for x1 in function(*args, **kwargs):
        for x2 in function(x1):
            for x3 in function(x2):
                out.append(x3)
    return out


class TestDefaultPipeline(BenchmarkCase):
    def setUp(self) -> None:
        self.pipeline = Step.chain(
            Step(function, loader='generator'),
            Step(function, loader='generator'),
            Step(function, loader='generator'),
        )

    def test_pipeline(self) -> None:
        baseline_time = timeit(30, baseline_pipeline, 'start', 15)
        pipeline_time = timeit(30, self.pipeline.run, 'start', 15)

        self.assertAlmostEqual(baseline_time, pipeline_time, places=2)



from abc import ABCMeta, abstractmethod
from functools import partial
from multiprocessing import Pool
from typing import Any, Callable, Iterable, List, Optional


class Loader(metaclass=ABCMeta):
    @property
    @abstractmethod
    def exactly_one_output(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def __call__(self, function: Callable, output: Any) -> Iterable[Any]:
        raise NotImplementedError()


class FunctionLoader(Loader):
    @property
    def exactly_one_output(self) -> bool:
        return True

    def __call__(self, function: Callable, output: Any) -> Iterable[Any]:
        yield function(output)


class GeneratorLoader(Loader):
    @property
    def exactly_one_output(self) -> bool:
        return False

    def __call__(self, function: Callable, output: Iterable[Any]) -> Iterable[Any]:
        yield from map(function, output)


class BatchLoader(GeneratorLoader):
    def __init__(self, batch_size: int, collate: Optional[Callable] = None) -> None:
        self._batch_size = batch_size
        self._collate = collate

    def collate(self, outputs: List[Any]) -> Any:
        if self._collate is None:
            return outputs
        return self._collate(outputs)

    def __call__(self, function: Callable, output: Iterable[Any]) -> Iterable[Any]:
        batch = []
        for raw_output in output:
            batch.append(raw_output)
            if len(batch) == self._batch_size:
                output = self.collate(batch)
                yield function(output)
                batch = []

        if batch:
            output = self.collate(batch)
            yield function(output)


class MultiprocessingLoader(GeneratorLoader):
    @staticmethod
    def function(function: Callable, output: Any) -> List[Any]:
        return list(function(output))

    def __call__(self, function: Callable, output: Iterable[Any]) -> Iterable[Any]:
        with Pool() as pool:
            yield from pool.imap(partial(self.function, function), output)



__version__ = '0.2.1'

from stepizer.loader import (
    BatchLoader, FunctionLoader, GeneratorLoader, Loader, MultiprocessingLoader)
from stepizer.proxy import DefaultProxy, MultipleProxy, PersistentProxy, Proxy
from stepizer.register import Register, loaders, proxies
from stepizer.step import Step

__all__ = [
    'BatchLoader',
    'DefaultProxy',
    'FunctionLoader',
    'GeneratorLoader',
    'Loader',
    'MultipleProxy',
    'MultiprocessingLoader',
    'PersistentProxy',
    'Proxy',
    'Register',
    'Step',
    'loaders',
    'proxies',
]



from typing import Iterable, List

from stepizer.step import Step
from tests.integration import IntegrationCase


def add_text(string: str) -> str:
    return 'hello, ' + string + '!'


class TestBasicPipeline(IntegrationCase):
    def setUp(self) -> None:
        self.pipeline = Step.chain(
            add_text,
            str.upper,
        )

    def test_pipeline(self) -> None:
        actual = self.pipeline.run('world')

        self.assertEqual('HELLO, WORLD!', actual)


def read_paths(directory: str) -> Iterable[str]:
    for i in range(3):
        yield f'{directory}/image_{i}.jpg'


def read_image(path: str) -> List[int]:
    return [ord(char) for char in path]


def detect_objects(image: List[int]) -> Iterable[List[int]]:
    for i in range(5, 7):
        yield image[-i:]


def save(obj: List[int], path: str) -> str:
    return path + f'.{sum(obj)}.save'


class TestAdvancedPipeline(IntegrationCase):
    def setUp(self) -> None:
        self.pipeline = (
            Step(read_paths, loader='generator', proxy_class='persistent')
            | read_image
            | Step(detect_objects, loader='generator')
            | Step(save, map_args=('', 'read_paths'))
        )

    def test_pipeline(self) -> None:
        expected = [
            'dir/image_0.jpg.415.save',
            'dir/image_0.jpg.510.save',
            'dir/image_1.jpg.416.save',
            'dir/image_1.jpg.511.save',
            'dir/image_2.jpg.417.save',
            'dir/image_2.jpg.512.save',
        ]

        actual = self.pipeline.run('dir')

        self.assertListEqual(expected, actual)



from typing import Dict, Generic, Iterator, Type, TypeVar, Union

from stepizer.loader import (
    BatchLoader, FunctionLoader, GeneratorLoader, Loader, MultiprocessingLoader)
from stepizer.proxy import DefaultProxy, MultipleProxy, PersistentProxy, Proxy

T = TypeVar('T')


class Register(Generic[T]):
    def __init__(self, **kwargs) -> None:
        super().__init__()
        self._items: Dict[str, Type[T]] = dict(**kwargs)

    def __getitem__(self, key: str) -> Type[T]:
        return self._items[key]

    def __setitem__(self, key: str, value: Type[T]) -> None:
        if key in self._items:
            raise KeyError(f"Cannot override {key} item")
        self._items[key] = value

    def __iter__(self) -> Iterator[str]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def init(self, _obj: Union[str, T], /, **kwargs) -> T:
        if isinstance(_obj, str):
            return self[_obj](**kwargs)  # type: ignore
        return _obj

    def get(self, _obj: Union[str, Type[T]], /) -> Type[T]:
        if isinstance(_obj, str):
            return self[_obj]
        return _obj


loaders: Register[Loader] = Register(
    batch=BatchLoader,
    function=FunctionLoader,
    generator=GeneratorLoader,
    multiprocessing=MultiprocessingLoader,
)

proxies: Register[Proxy] = Register(
    default=DefaultProxy,
    persistent=PersistentProxy,
    multiple=MultipleProxy,
)



from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterable, Optional

if typing.TYPE_CHECKING:
    from stepizer.step import Step


class Proxy(metaclass=ABCMeta):
    def __init__(self, step: Step, cache: Optional[Dict[str, Any]] = None) -> None:
        self._step = step
        self._cache = cache or dict()

    def __call__(self, output: Any) -> Iterable[Any]:
        if self._step.next_step is None:
            yield output
        else:
            cache = {**self._cache, **self.update_cache(output)}
            _cache = {**cache, '': output}

            args = tuple(_cache[name] for name in self._step.next_step.map_args)
            kwargs = {key: _cache[name] for key, name in self._step.next_step.map_kwargs.items()}

            yield from self._step.next_step(*args, _cache=cache, **kwargs)

    @abstractmethod
    def update_cache(self, output: Any) -> Dict[str, Any]:
        raise NotImplementedError


class DefaultProxy(Proxy):
    def update_cache(self, output: Any) -> Dict[str, Any]:
        return dict()


class PersistentProxy(Proxy):
    def update_cache(self, output: Any) -> Dict[str, Any]:
        return {self._step.name: output}


class MultipleProxy(Proxy):
    def update_cache(self, output: Any) -> Dict[str, Any]:
        return output



from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Type, Union

from stepizer.loader import Loader
from stepizer.proxy import Proxy
from stepizer.register import loaders, proxies


class Step:
    def __init__(
        self,
        _callable: Callable,
        /,
        args: Tuple[Any, ...] = tuple(),
        kwargs: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        loader: Union[Loader, str] = 'function',
        loader_kwargs: Optional[Dict[str, Any]] = None,
        proxy_class: Union[Type[Proxy], str] = 'default',
        map_args: Tuple[str, ...] = ('',),
        map_kwargs: Optional[Dict[str, str]] = None,
    ) -> None:
        self._callable = _callable
        self._args = args
        self._kwargs = kwargs or dict()
        self._name = name or self._callable.__name__
        self._loader = loaders.init(loader, **(loader_kwargs or dict()))
        self._proxy_class = proxies.get(proxy_class)
        self._map_args = map_args
        self._map_kwargs = map_kwargs or dict()
        self._next_step: Optional[Step] = None

    @property
    def callable(self) -> Callable:
        return self._callable

    @property
    def name(self) -> str:
        return self._name

    @property
    def loader(self) -> Loader:
        return self._loader

    @property
    def proxy_class(self) -> Type[Proxy]:
        return self._proxy_class

    @property
    def map_args(self) -> Tuple[str, ...]:
        return self._map_args

    @property
    def map_kwargs(self) -> Dict[str, str]:
        return self._map_kwargs

    @property
    def next_step(self) -> Optional[Step]:
        return self._next_step

    def iter_steps(self) -> Iterable[Step]:
        yield self
        if self._next_step:
            yield from self._next_step.iter_steps()

    @classmethod
    def wrap(cls, _callable: Callable, /) -> Step:
        if isinstance(_callable, cls):
            return _callable
        return cls(_callable)

    @classmethod
    def chain(cls, _callable: Callable, /, *callables: Callable) -> Step:
        step = cls.wrap(_callable)
        for c in callables:
            step.link(c)
        return step

    def link(self, _callable: Callable, /) -> Step:
        if self._next_step is None:
            self._next_step = Step.wrap(_callable)
        else:
            self._next_step.link(_callable)
        return self

    __or__ = link

    def __call__(self, *args, _cache: Optional[Dict[str, Any]] = None, **kwargs) -> Iterable[Any]:
        proxy = self._proxy_class(step=self, cache=_cache)
        output = self._callable(*self._args, *args, **self._kwargs, **kwargs)
        for output in self._loader(function=proxy, output=output):
            yield from output

    def run(self, *args, **kwargs) -> Any:
        output = self(*args, **kwargs)
        if all(step._loader.exactly_one_output for step in self.iter_steps()):
            return next(iter(output))
        return list(output)

    def execute(self, *args, **kwargs) -> None:
        for _ in self(*args, **kwargs):
            pass
